/*
 * CSRFGuard.java
 *
 * Created on January 2, 2007, 11:35 AM
 *
 * Copyright (C) 2007 Eric Sheridan
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package org.owasp.csrf;

import java.io.OutputStream;
import java.io.IOException;
import java.security.SecureRandom;
import java.util.regex.Pattern;
import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpSession;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import sun.misc.BASE64Encoder;
import org.owasp.csrf.http.MutableHttpResponse;

/**
 *
 * @author esheridan
 */
public class CSRFGuard implements Filter {
    public final static String OWASP_CSRFTOKEN = "OWASP_CSRFTOKEN";
    
    public final static Pattern FORM_PATTERN = Pattern.compile("(?i)</form>");
    
    public final static Pattern SKIPPABLE_PATTERN = Pattern.compile(".*\\.(gif|jpg|png|css|js|ico|swf|axd.*)$");
    
    private String errorPage = null;
    
    public void init(FilterConfig config) {
        errorPage = config.getInitParameter("error-page");
    }
    
    public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain) {
        if(request instanceof HttpServletRequest && response instanceof HttpServletResponse) {
            try {
                HttpServletRequest hRequest = (HttpServletRequest)request;
                MutableHttpResponse mResponse = new MutableHttpResponse((HttpServletResponse)response);
                HttpSession session = hRequest.getSession(false);
                
                if(session != null) {
                    if(!hasCSRFToken(session)) {
                        setCSRFToken(session);
                        doChain(hRequest, mResponse, response, chain);
                    } else if(isSkippable(hRequest) || isValidRequest(session, hRequest)) {
                        doChain(hRequest, mResponse, response, chain);
                    } else {
                        doError(session, (HttpServletResponse)response);
                    }
                } else {
                    chain.doFilter(request, response);
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }
    
    public void destroy() {
        
    }
    
    private void doError(HttpSession session, HttpServletResponse response) throws IOException {
        session.invalidate();
        response.sendRedirect(errorPage);
    }
    
    private void doChain(HttpServletRequest request, MutableHttpResponse mResponse, ServletResponse response, FilterChain chain) throws IOException, ServletException {
        OutputStream out = response.getOutputStream();
        
        chain.doFilter(request, mResponse);
        
        String contentType = mResponse.getContentType();
        String token = getCSRFToken(request.getSession());
        
        if(contentType != null && contentType.startsWith("text")) {
            String content = new String(mResponse.getContent());
            content = getModifiedResponse(token, content);
            byte[] result = content.getBytes();
            
            response.setContentLength(result.length);
            out.write(result);
        } else {
            out.write(mResponse.getContent());
        }
        
        out.close();
    }
    
    private String getModifiedResponse(String token, String content) {
        content = buildFormParameters(content, token);
        content = buildLinkParameters(content, token);
        
        return content;
    }
    
    private String getCSRFToken(HttpSession session) {
        return (String)session.getAttribute(OWASP_CSRFTOKEN);
    }
    
    private boolean hasCSRFToken(HttpSession session) {
        return getCSRFToken(session) != null;
    }
    
    private void setCSRFToken(HttpSession session) {
        session.setAttribute(OWASP_CSRFTOKEN, generateCSRFToken());
    }
    
    private String generateCSRFToken() {
        return generateCSRFToken(32);
    }
    
    private String generateCSRFToken(int size) {
        SecureRandom sr = null;
        byte[] random = new byte[size];
        BASE64Encoder encoder = new BASE64Encoder();
        String digest = null;
        
        try {
            sr = SecureRandom.getInstance("SHA1PRNG");
            
            sr.nextBytes(random);
            
            digest = encoder.encode(random);
        } catch (Exception e) {
            e.printStackTrace();
        }
        
        return digest;
    }
    
    private boolean isSkippable(HttpServletRequest request) {
        return SKIPPABLE_PATTERN.matcher(request.getRequestURI()).matches();
    }
    
    private boolean isValidRequest(HttpSession session, HttpServletRequest request) {
        String original = (String)session.getAttribute(OWASP_CSRFTOKEN);
        String now = (String)request.getParameter(OWASP_CSRFTOKEN);
        boolean result = false;
        
        if(now != null && now.equals(original)) {
            result = true;
        }
        
        return result;
    }
    
    private String buildFormParameters(String content, String token) {
        return FORM_PATTERN.matcher(content).replaceAll("<input type=\"hidden\" name=\"" + OWASP_CSRFTOKEN + "\" value=\"" + token + "\">\n</form>");
    }
    
    private String buildLinkParameters(String content, String token) {
        content = buildLinkParameters(content, token, "href=\"", '"');
        content = buildLinkParameters(content, token, "src=\"", '"');
        
        return content;
    }
    
    private String buildLinkParameters(String content, String token, String pattern, char termChar) {
        StringBuffer buffer = new StringBuffer();
        int i=0;
        int index = 0;
        int length = content.length();
        
        while(i < length) {
            index = content.toLowerCase().indexOf(pattern, i);
            
            if(index != -1) {
                int offset = 0;
                int n = 0;
                boolean parameters = false;
                String tokenString = null;
                buffer.append(content.substring(i, index+pattern.length()));
                
                for(n=index+pattern.length(), offset=n; n<=length && content.charAt(n) != termChar; n++) {
                    buffer.append(content.charAt(n));
                    
                    if(content.charAt(n) == '?') {
                        parameters = true;
                    }
                }
                
                if(parameters) {
                    tokenString = "&" + OWASP_CSRFTOKEN + "=" + token;
                } else {
                    tokenString = "?" + OWASP_CSRFTOKEN + "=" + token;
                }
                
                buffer.append(tokenString);
                
                i = index + pattern.length() + (n - offset);
            } else {
                buffer.append(content.substring(i, length));
                
                i = length;
            }
        }
        
        return buffer.toString();
    }
}

